from functools import partial, wraps

import click


# TODO - unsure if this is the right idiom
def catch_exception(func=None, *, handle):
    """catch exception decorator"""
    if not func:
        return partial(catch_exception, handle=handle)

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except handle as e:
            raise click.ClickException(e)

    return wrapper



import pytest
from click.testing import CliRunner

from griffon.cli import cli

pytestmark = pytest.mark.unit


def test_osv():
    runner = CliRunner()
    result = runner.invoke(cli, ["plugins", "osv"])
    assert result.exit_code == 0


def test_fcc():
    runner = CliRunner()
    result = runner.invoke(cli, ["plugins", "fcc"])
    assert result.exit_code == 0



"""

"""

import click

from griffon import CORGI_API_URL, OSIDB_API_URL

# https://docs.engineering.redhat.com/display/PRODSEC/Product+Security+Services+and+Tools+Catalog
# https://docs.engineering.redhat.com/pages/viewpage.action?pageId=81867704

# https://dashboard.prodsec.redhat.com/
# https://git.prodsec.redhat.com/
# http://opengrok.brq.redhat.com/source/
# https://errata.engineering.redhat.com/errata
# https://pp.engineering.redhat.com/pp/
# https://access.redhat.com/downloads/content/package-browser


@click.group(name="docs", help="Links to useful docs.")
@click.pass_context
def docs_grp(ctx):
    pass


@docs_grp.group(name="service", help="Links to coarse grained data services.")
@click.pass_context
def service_grp(ctx):
    pass


@service_grp.command()
def griffon_github():
    click.launch("https://github.com/RedHatProductSecurity/griffon")


@service_grp.command()
def griffon_tutorial():
    click.launch("https://github.com/RedHatProductSecurity/griffon/blob/main/docs/tutorial.md")


@service_grp.command()
def osidb():
    click.launch(OSIDB_API_URL)


@service_grp.command()
def osidb_github():
    click.launch("https://github.com/RedHatProductSecurity/osidb")


@service_grp.command()
def osidb_tutorial():
    click.launch("https://github.com/RedHatProductSecurity/osidb/blob/master/docs/user/TUTORIAL.md")


@service_grp.command()
def osidb_bindings():
    click.launch("https://github.com/RedHatProductSecurity/osidb-bindings/blob/master/TUTORIAL.md")


@service_grp.command()
def corgi():
    click.launch(CORGI_API_URL)


@service_grp.command()
def corgi_github():
    click.launch("https://github.com/RedHatProductSecurity/component-registry")


@service_grp.command()
def corgi_tutorial():
    click.launch(
        "https://github.com/RedHatProductSecurity/component-registry/blob/main/docs/user_guide.md"
    )


@service_grp.command()
def package_browser(help="Red Hat own package browser"):
    click.launch("https://access.redhat.com/downloads/content/package-browser")



"""

"""
import configparser
import io
import logging
import os
import subprocess

import click

from griffon import GRIFFON_CONFIG_DIR, GRIFFON_DEFAULT_LOG_FILE, GRIFFON_RC_FILE

logger = logging.getLogger("griffon")


@click.group(name="configure")
@click.pass_context
def configure_grp(ctx):
    """Configure griffon."""
    pass


@configure_grp.command(help="Update griffon to latest release")
@click.pass_context
def update(ctx):
    subprocess.run(["pip", "install", "--force", "griffon"])


@configure_grp.command(name="setup", help="Create ~/.griffon and .griffonrc config file")
def setup():
    """stub"""
    if not os.path.exists(os.path.expanduser(GRIFFON_CONFIG_DIR)):
        os.makedirs(os.path.expanduser(GRIFFON_CONFIG_DIR))
        logger.warning(f"{GRIFFON_CONFIG_DIR} created")
    else:
        logger.warning(f"{GRIFFON_CONFIG_DIR} already exists")

    with open("griffon/static/default_griffonrc", "r") as file:
        data = file.read()
    config = configparser.ConfigParser(allow_no_value=True)
    config.readfp(io.StringIO(data))
    config.set("default", "history_log", GRIFFON_DEFAULT_LOG_FILE)

    with open(os.path.expanduser(GRIFFON_RC_FILE), "w") as configfile:
        config.write(configfile)



"""
    read only queries

"""
import concurrent
import logging
from datetime import datetime

import requests

from griffon import OSIDB_API_URL, CorgiService, OSIDBService

logger = logging.getLogger("griffon")


class example_affects_report:
    """ """

    name = "example_affects_report"
    description = " "
    allowed_params = [
        "product_version_name",
        "show_components",
        "all",
        "show_products",
        "purl",
        "name",
        "ofuri",
        "product_name",
    ]

    def __init__(self, params) -> None:
        self.corgi_session = CorgiService.create_session()
        self.osidb_session = OSIDBService.create_session()
        self.params = params
        self.product_version_name = self.params.get("product_version_name")
        self.show_components = self.params.get("show_components")
        self.show_products = self.params.get("show_products")
        self.all = self.params.get("all")
        self.purl = self.params.get("purl")
        self.component_name = self.params.get("name")
        self.ofuri = self.params.get("ofuri")
        self.product_name = self.params.get("product_name")

    def generate(self) -> dict:
        affects = self.osidb_session.affects.retrieve_list()

        affects_affected = self.osidb_session.affects.retrieve_list(affectedness="AFFECTED")

        # TODO osidb_session.affects does not support include_fields, do it manually for now
        affect_critical_res = requests.get(
            f"{OSIDB_API_URL}/osidb/api/v1/affects?affectedness=AFFECTED&impact=CRITICAL&include_fields=ps_component,ps_module,impact&limit=10000"  # noqa
        )
        affects_critical = affect_critical_res.json()
        critical_components = {}
        critical_products = {}
        for affect in affects_critical["results"]:
            component_name = affect.get("ps_component", "no_name")
            if component_name not in critical_components:
                critical_components[component_name] = 0
            critical_components[component_name] += 1

            product_version_name = affect.get("ps_module", "no_name")
            if product_version_name not in critical_products:
                critical_products[product_version_name] = 0
            critical_products[product_version_name] += 1

        top_critical_component = max(critical_components, key=critical_components.get)  # type: ignore # noqa
        top_critical_product = max(critical_products, key=critical_products.get)  # type: ignore

        affect_important_res = requests.get(
            f"{OSIDB_API_URL}/osidb/api/v1/affects?affectedness=AFFECTED&impact=IMPORTANT&include_fields=ps_component,ps_module,impact&limit=10000"  # noqa
        )
        affects_important = affect_important_res.json()
        important_components = {}
        important_products = {}
        for affect in affects_important["results"]:
            component_name = affect.get("ps_component", "no_name")
            if component_name not in important_components:
                important_components[component_name] = 0
            important_components[component_name] += 1

            product_version_name = affect.get("ps_module", "no_name")
            if product_version_name not in important_products:
                important_products[product_version_name] = 0
            important_products[product_version_name] += 1
        top_important_component = max(important_components, key=important_components.get)  # type: ignore # noqa
        top_important_product = max(important_products, key=important_products.get)  # type: ignore # noqa

        affect_moderate_res = requests.get(
            f"{OSIDB_API_URL}/osidb/api/v1/affects?affectedness=AFFECTED&impact=MODERATE&include_fields=ps_component,ps_module,impact&limit=10000"  # noqa
        )
        affects_moderate = affect_moderate_res.json()
        moderate_components = {}
        moderate_products = {}
        for affect in affects_moderate["results"]:
            component_name = affect.get("ps_component", "no_name")
            if component_name not in moderate_components:
                moderate_components[component_name] = 0
            moderate_components[component_name] += 1

            product_version_name = affect.get("ps_module", "no_name")
            if product_version_name not in moderate_products:
                moderate_products[product_version_name] = 0
            moderate_products[product_version_name] += 1
        top_moderate_component = max(moderate_components, key=moderate_components.get)  # type: ignore # noqa
        top_moderate_product = max(moderate_products, key=moderate_products.get)  # type: ignore

        affect_low_res = requests.get(
            f"{OSIDB_API_URL}/osidb/api/v1/affects?affectedness=AFFECTED&impact=LOW&include_fields=ps_component,ps_module,impact&limit=10000"  # noqa
        )
        affects_low = affect_low_res.json()
        low_components = {}
        low_products = {}
        for affect in affects_low["results"]:
            component_name = affect.get("ps_component", "no_name")
            if component_name not in low_components:
                low_components[component_name] = 0
            low_components[component_name] += 1

            product_version_name = affect.get("ps_module", "no_name")
            if product_version_name not in low_products:
                low_products[product_version_name] = 0
            low_products[product_version_name] += 1
        top_low_component = max(low_components, key=low_components.get)  # type: ignore
        top_low_product = max(low_products, key=low_products.get)  # type: ignore

        critical = {
            "affected": affects_critical["count"],
            "top_component": top_critical_component,
            "top_product": top_critical_product,
        }
        important = {
            "affected": affects_important["count"],
            "top_component": top_important_component,
            "top_product": top_important_product,
        }
        moderate = {
            "affected": affects_moderate["count"],
            "top_component": top_moderate_component,
            "top_product": top_moderate_product,
        }
        low = {
            "affected": affects_low["count"],
            "top_component": top_low_component,
            "top_product": top_low_product,
        }

        if self.show_components:
            critical["components"] = critical_components
            important["components"] = important_components
            moderate["components"] = moderate_components
            low["components"] = low_components

        if self.show_products:
            critical["products"] = critical_products
            important["products"] = important_products
            moderate["products"] = moderate_products
            low["products"] = low_products

        report = {
            "title": "Example Affects report",
            "ts": str(datetime.now()),
            "total_affects": affects.count,
            "total_affected": affects_affected.count,
            "critical": critical,
            "important": important,
            "moderate": moderate,
            "low": low,
        }

        if self.purl:
            report["purl"] = self.purl
        if self.component_name:
            report["component_name"] = self.component_name
        if self.ofuri:
            report["product_ofuri"] = self.ofuri
        if self.product_name:
            report["product"] = self.product_name

        return {
            "title": "Example Affects report",
            "ts": str(datetime.now()),
            "total_affects": affects.count,
            "total_affected": affects_affected.count,
            "critical": critical,
            "important": important,
            "moderate": moderate,
            "low": low,
        }


class entity_report:
    """ """

    name = "entity_report"
    description = " "
    allowed_params = ["all"]

    def __init__(self, params) -> None:
        self.corgi_session = CorgiService.create_session()
        self.osidb_session = OSIDBService.create_session()
        self.params = params

    def generate(self) -> dict:
        component_arches = CorgiService.get_component_arches()
        component_types = [
            component_type.value for component_type in CorgiService.get_component_types()
        ]
        corgi_status = self.corgi_session.status()
        active_product_streams_count = self.corgi_session.product_streams.retrieve_list().count
        component_instances_count = self.corgi_session.components.retrieve_list(limit=1).count

        # get src RPM count
        component_cnt = self.corgi_session.components.retrieve_list(
            arch="src", namespace="REDHAT", type="RPM", include_fields="name"
        ).count
        components = list()
        if component_cnt < 3000000:
            with concurrent.futures.ThreadPoolExecutor() as executor:
                futures = []
                for batch in range(0, component_cnt, 5000):
                    futures.append(
                        executor.submit(
                            self.corgi_session.components.retrieve_list,
                            arch="src",
                            namespace="REDHAT",
                            type="RPM",
                            include_fields="name",
                            offset=batch,
                            limit=5000,  # noqa
                        )
                    )

                for future in concurrent.futures.as_completed(futures):
                    try:
                        components.extend(future.result().results)
                    except Exception as exc:
                        logger.warning("%r generated an exception: %s" % (future, exc))

        # # get OCI noarch count
        # component_cnt = self.corgi_session.components.retrieve_list(
        #     arch="noarch", type="OCI", include_fields="name"
        # ).count
        # oci_components = list()
        # if component_cnt < 3000000:
        #     with concurrent.futures.ThreadPoolExecutor() as executor:
        #         futures = []
        #         for batch in range(0, component_cnt, 3000):
        #             futures.append(
        #                 executor.submit(
        #                     self.corgi_session.components.retrieve_list,
        #                     arch="noarch",
        #                     type="OCI",
        #                     include_fields="name",
        #                     offset=batch,
        #                     limit=3000,  # noqa
        #                 )
        #             )
        #
        #         for future in concurrent.futures.as_completed(futures):
        #             try:
        #                 oci_components.extend(future.result().results)
        #             except Exception as exc:
        #                 logger.warning("%r generated an exception: %s" % (future, exc))

        # # get NPM noarch count
        # component_cnt = self.corgi_session.components.retrieve_list(type="NPM",
        #                                                             include_fields="name").count
        # npm_components = list()
        # if component_cnt < 3000000:
        #     with concurrent.futures.ThreadPoolExecutor() as executor:
        #         futures = []
        #         for batch in range(0, component_cnt, 3000):
        #             futures.append(
        #                 executor.submit(
        #                     self.corgi_session.components.retrieve_list,
        #                     type="NPM", include_fields="name",
        #                     offset=batch,
        #                     limit=3000,  # noqa
        #                 )
        #             )
        #
        #         for future in concurrent.futures.as_completed(futures):
        #             try:
        #                 npm_components.extend(future.result().results)
        #             except Exception as exc:
        #                 logger.warning("%r generated an exception: %s" % (future, exc))

        # get GOLANG noarch count
        # component_cnt = self.corgi_session.components.retrieve_list(type="GOLANG",
        #                                                             include_fields="name").count
        # golang_components = list()
        # if component_cnt < 3000000:
        #     with concurrent.futures.ThreadPoolExecutor() as executor:
        #         futures = []
        #         for batch in range(0, component_cnt, 3000):
        #             futures.append(
        #                 executor.submit(
        #                     self.corgi_session.components.retrieve_list,
        #                     type="GOLANG", include_fields="name",
        #                     offset=batch,
        #                     limit=3000,  # noqa
        #                 )
        #             )
        #
        #         for future in concurrent.futures.as_completed(futures):
        #             try:
        #                 golang_components.extend(future.result().results)
        #             except Exception as exc:
        #                 logger.warning("%r generated an exception: %s" % (future, exc))

        rpm_components_cnt: int = len(list(set([component.name for component in components])))

        # TODO: baking these values in as we will eventually process these server side
        oci_components_cnt: int = 2873
        npm_components_cnt: int = 8677
        golang_components_cnt: int = 48359

        total_component_cnt = (
            rpm_components_cnt + oci_components_cnt + npm_components_cnt + golang_components_cnt
        )

        return {
            "corgi": {
                "title": "Entity report",
                "ts": str(datetime.now()),
                "db_size": corgi_status["db_size"],
                "components": {
                    "types": component_types,
                    "arches": component_arches,
                    "total_component_instances": component_instances_count,
                    "total_distinct_components": total_component_cnt,
                    "rpm_components": rpm_components_cnt,
                    "oci_components": oci_components_cnt,
                    "npm_components": npm_components_cnt,
                    "golang_components": golang_components_cnt,
                },
                "products": {
                    "products": corgi_status["products"]["count"],
                    "product_versions": corgi_status["product_versions"]["count"],
                    "active_product_streams": active_product_streams_count,
                    "product_streams": corgi_status["product_streams"]["count"],
                    "product_variants": corgi_status["product_variants"]["count"],
                    "channels": corgi_status["channels"]["count"],
                },
            }
        }



"""
entity operations

"""
import concurrent.futures
import logging

import click

from griffon import (
    CORGI_API_URL,
    OSIDB_API_URL,
    CorgiService,
    OSIDBService,
    progress_bar,
)
from griffon.autocomplete import (
    get_component_names,
    get_component_purls,
    get_cve_ids,
    get_product_stream_names,
    get_product_stream_ofuris,
)
from griffon.output import console, cprint

logger = logging.getLogger("griffon")

default_conditions: dict = {}


@click.group(name="entities", help="Entity operations (UNDER DEVELOPMENT).")
@click.option("--open-browser", is_flag=True, help="open browser to service results.")
@click.option("--limit", default=10, help="# of items returned by list operations.")
@click.pass_context
def entities_grp(ctx, open_browser, limit):
    ctx.ensure_object(dict)
    ctx.obj["open_browser"] = open_browser
    ctx.obj["limit"] = limit


# flaws
@entities_grp.group(help=f"{OSIDB_API_URL}/osidb/api/v1/flaws")
@click.pass_context
def flaws(ctx):
    """OSIDB Flaws."""


@flaws.command(name="list")
@click.option(
    "--state",
    "flaw_state",
    type=click.Choice(OSIDBService.get_flaw_states()),
    help="Flaw state.",
)
@click.option(
    "--resolution",
    type=click.Choice(OSIDBService.get_flaw_resolutions()),
    help="Flaw resolution.",
)
@click.option(
    "--impact",
    type=click.Choice(OSIDBService.get_flaw_impacts()),
    help="Flaw impact.",
)
@click.option(
    "--embargoed",
    "is_embargoed",
    is_flag=True,
    help="Include embargoed flaws (requires access).",
)
@click.option("--major-incident", "is_major_incident", is_flag=True, help="Only major incidents.")
@click.pass_context
@progress_bar
def list_flaws(ctx, flaw_state, resolution, impact, is_embargoed, is_major_incident):
    if not flaw_state and not resolution and not impact:
        click.echo(ctx.get_help())
        exit(0)
    session = OSIDBService.create_session()
    conditions = default_conditions
    if flaw_state:
        conditions["state"] = flaw_state
    if resolution:
        conditions["resolution"] = resolution
    if impact:
        conditions["impact"] = impact
    data = session.flaws.retrieve_list(**conditions).results
    return cprint(data, ctx=ctx)


@flaws.command(name="get")
@click.option("--cve-id", help="Flaw CVE-ID.", shell_complete=get_cve_ids)
@click.option(
    "--uuid",
    "flaw_uuid",
    help="Flaw UUID.",
)
@click.pass_context
def get_flaw(ctx, cve_id, flaw_uuid):
    if not cve_id and not flaw_uuid:
        click.echo(ctx.get_help())
        exit(0)
    session = OSIDBService.create_session()
    if flaw_uuid:
        data = session.flaws.retrieve(flaw_uuid)
    if cve_id:
        data = session.flaws.retrieve(cve_id)
    return cprint(data, ctx=ctx)


# affects
@entities_grp.group(help=f"{OSIDB_API_URL}/osidb/api/v1/affects")
@click.pass_context
def affects(ctx):
    """OSIDB Affects."""
    pass


@affects.command(name="list")
@click.option("--product_version", help="ps module")
@click.option("--component_name", help="ps component")
@click.option("--affectedness", type=click.Choice(OSIDBService.get_affect_affectedness()))
@click.option(
    "--resolution",
    type=click.Choice(OSIDBService.get_affect_resolution()),
)
@click.option("--impact", type=click.Choice(OSIDBService.get_affect_impact()))
@click.pass_context
@progress_bar
def list_affects(ctx, product_version, component_name, affectedness, resolution, impact):
    if (
        not product_version
        and not component_name
        and not affectedness
        and not resolution
        and not impact
    ):
        click.echo(ctx.get_help())
        exit(0)
    session = OSIDBService.create_session()
    conditions = default_conditions
    if product_version:
        conditions["ps_module"] = product_version
    if component_name:
        conditions["ps_component"] = component_name
    if affectedness:
        conditions["affectedness"] = affectedness
    if resolution:
        conditions["resolution"] = resolution
    if impact:
        conditions["impact"] = impact
    data = session.affects.retrieve_list(**conditions).results
    return cprint(data, ctx=ctx)


@affects.command(name="get")
@click.option("--uuid", "affect_uuid")
@click.pass_context
def get_affect(ctx, affect_uuid):
    if not affect_uuid:
        click.echo(ctx.get_help())
        exit(0)
    session = OSIDBService.create_session()
    data = session.affects.retrieve(affect_uuid)
    return cprint(data, ctx=ctx)


# trackers
@entities_grp.group(help=f"{OSIDB_API_URL}/osidb/api/v1/trackers")
@click.pass_context
def trackers(ctx):
    """OSIDB Trackers."""
    pass


@trackers.command(name="list")
@click.pass_context
def list_trackers(ctx):
    session = OSIDBService.create_session()
    conditions = default_conditions
    data = session.trackers.retrieve_list(**conditions).results
    return cprint(data, ctx=ctx)


@trackers.command(name="get")
@click.option("--uuid", "tracker_uuid")
@click.pass_context
@progress_bar
def get_tracker(ctx, tracker_uuid):
    if not tracker_uuid:
        click.echo(ctx.get_help())
        exit(0)
    session = OSIDBService.create_session()
    data = session.trackers.retrieve(tracker_uuid)
    return cprint(data, ctx=ctx)


# components
@entities_grp.group(help=f"{CORGI_API_URL}/api/v1/components")
@click.pass_context
def components(ctx):
    pass


@components.command(name="list")
@click.argument("component_name", required=False)
@click.option("--namespace", type=click.Choice(CorgiService.get_component_namespaces()), help="")
@click.option("--ofuri", shell_complete=get_product_stream_ofuris)
@click.option("--re_purl", shell_complete=get_component_purls)
@click.option("--re_name", shell_complete=get_component_names)
@click.option("--version")
@click.option(
    "--type",
    "component_type",
    type=click.Choice(CorgiService.get_component_types()),
    help="",  # noqa
)
@click.option(
    "--arch",
    type=click.Choice(CorgiService.get_component_arches()),
)
@click.option("--product-stream-name", shell_complete=get_product_stream_names)
@click.option("--product-stream-ofuri", shell_complete=get_product_stream_ofuris)
@click.pass_context
@progress_bar
def list_components(
    ctx,
    component_name,
    namespace,
    ofuri,
    re_purl,
    re_name,
    version,
    component_type,
    arch,
    product_stream_name,
    product_stream_ofuri,
):
    """Retrieve a list of Components."""

    if (
        not component_name
        and not ofuri
        and not re_purl
        and not re_name
        and not version
        and not arch
        and not namespace
        and not component_type
        and not product_stream_name
        and not product_stream_ofuri
    ):
        click.echo(ctx.get_help())
        exit(0)
    session = CorgiService.create_session()

    conditions = default_conditions
    conditions["include_fields"] = "link,purl,nvr,version,type,name,upstreams,related_url"

    # TODO- condition union could be a separate helper function

    if component_name:
        conditions["name"] = component_name

    if namespace:
        conditions["namespace"] = namespace
    if ofuri:
        conditions["ofuri"] = ofuri
    if re_purl:
        conditions["re_purl"] = re_purl
    if re_name:
        conditions["re_name"] = re_name
    if version:
        conditions["version"] = version
    if arch:
        conditions["arch"] = arch
    if component_type:
        conditions["type"] = component_type
    if product_stream_ofuri:
        conditions["product_streams"] = product_stream_ofuri

    # TODO- This kind of optimisation should probably be developed in the
    #       service binding itself rather then here
    logger.debug("starting parallel http requests")
    component_cnt = session.components.retrieve_list(**conditions).count
    if component_cnt < 3000000:
        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = []
            components = list()
            for batch in range(0, component_cnt, 1200):
                futures.append(
                    executor.submit(
                        session.components.retrieve_list,
                        **conditions,
                        offset=batch,
                        limit=1200,  # noqa
                    )
                )

            for future in concurrent.futures.as_completed(futures):
                try:
                    components.extend(future.result().results)
                except Exception as exc:
                    logger.warning("%r generated an exception: %s" % (future, exc))

            data = sorted(components, key=lambda d: d.purl)
            return cprint(data, ctx=ctx)
    else:
        console.print("downloading too many")


@components.command(name="get")
@click.option("--uuid", "component_uuid")
@click.option("--purl", shell_complete=get_component_purls, help="Purl are URI and must be quoted.")
@click.option("--nvr")
@click.pass_context
@progress_bar
def get_component(ctx, component_uuid, purl, nvr):
    """Retrieve Component."""
    if not component_uuid and not purl and not nvr:
        click.echo(ctx.get_help())
        exit(0)
    session = CorgiService.create_session()
    data = session.components.retrieve_list(purl=purl)
    return cprint(data, ctx=ctx)


@components.command(
    name="summary",
    help="Get Component summaries.",
)
@click.argument(
    "component_name",
    required=False,
)
@click.option(
    "-s",
    "strict_name_search",
    is_flag=True,
    default=False,
    help="Strict search, exact match of name.",
)
@click.pass_context
@progress_bar
def get_component_summary(ctx, component_name, strict_name_search):
    """Get Component summary."""
    if not component_name:
        click.echo(ctx.get_help())
        exit(0)
    session = CorgiService.create_session()

    cond = {
        "include_fields": "name,purl,tags,arch,release,version,product_streams,upstreams,related_url",  # noqa
        "name": component_name,
    }
    components = session.components.retrieve_list(**cond, limit=10000)
    product_streams = []
    upstreams = []
    versions = []
    releases = []
    arches = []
    related_urls = []
    tags = []
    for component in components.results:
        related_urls.append(component.related_url)
        arches.append(component.arch)
        versions.append(component.version)
        releases.append(component.release)
        tags.extend(component.tags)
        for upstream in component.upstreams:
            upstreams.append(upstream["purl"])
        for ps in component.product_streams:
            product_streams.append(ps["name"])
    data = {
        "link": f"{CORGI_API_URL}/api/v1/components?name={component_name}",
        "name": component_name,
        "tags": sorted(list(set(tags))),
        "count": len(components.results),
        "product_streams": sorted(list(set(product_streams))),
        "related_urls": sorted(list(set(related_urls))),
        "releases": sorted(list(set(releases))),
        "upstreams": sorted(list(set(upstreams))),
        "arches": sorted(list(set(arches))),
        "versions": sorted(list(set(versions))),
    }
    cprint(data, ctx=ctx)


@components.command(name="provides")
@click.option("--uuid", "component_uuid")
@click.option("--purl", shell_complete=get_component_purls, help="Purl are URI and must be quoted.")
@click.pass_context
@progress_bar
def get_component_provides(ctx, component_uuid, purl):
    """Retrieve all Components provided by a Component."""
    if not component_uuid and not purl:
        click.echo(ctx.get_help())
        exit(0)
    session = CorgiService.create_session()
    if component_uuid:
        data = session.components.retrieve(component_uuid).provides
        return cprint(data, ctx=ctx)
    else:
        c = session.components.retrieve_list(purl=purl)
        if c:
            data = session.components.retrieve(c["uuid"]).provides
            return cprint(data, ctx=ctx)


@components.command(name="sources")
@click.option("--uuid", "component_uuid")
@click.option("--purl", shell_complete=get_component_purls, help="Purl are URI and must be quoted.")
@click.pass_context
@progress_bar
def get_component_sources(ctx, component_uuid, purl):
    """Retrieve all Components that contain Component."""
    if not component_uuid and not purl:
        click.echo(ctx.get_help())
        exit(0)
    session = CorgiService.create_session()
    if component_uuid:
        data = session.components.retrieve(component_uuid).sources
        return cprint(data, ctx=ctx)
    else:
        c = session.components.retrieve_list(purl=purl)
        if c:
            data = session.components.retrieve(c["uuid"]).sources
            return cprint(data, ctx=ctx)


@components.command(name="manifest")
@click.option("--uuid", "component_uuid")
@click.option("--purl", shell_complete=get_component_purls, help="Purl are URI and must be quoted.")
@click.pass_context
@progress_bar
def get_component_manifest(ctx, component_uuid, purl):
    """Retrieve Component manifest."""
    if not component_uuid and not purl:
        click.echo(ctx.get_help())
        exit(0)
    session = CorgiService.create_session()
    if component_uuid:
        data = session.components.retrieve_manifest(component_uuid)
        return cprint(data, ctx=ctx)
    else:
        c = session.components.retrieve_list(purl=purl)
        if c:
            data = session.components.retrieve_manifest(c["uuid"])
            return cprint(data, ctx=ctx)


# product streams
@entities_grp.group(help=f"{CORGI_API_URL}/api/v1/product_streams")
@click.pass_context
def product_streams(ctx):
    pass


@product_streams.command(name="list")
@click.argument(
    "product_stream_name",
    required=False,
    type=click.STRING,
    shell_complete=get_product_stream_names,
)
@click.pass_context
def list_product_streams(ctx, product_stream_name):
    """Retrieve a list of Product Streams."""
    session = CorgiService.create_session()
    cond = default_conditions
    if product_stream_name:
        cond["re_name"] = product_stream_name
    data = session.product_streams.retrieve_list(**cond, limit=1000).results
    return cprint(data, ctx=ctx)


@product_streams.command(name="get")
@click.argument(
    "product_stream_name",
    required=False,
    type=click.STRING,
    shell_complete=get_product_stream_names,
)
@click.option("--inactive", is_flag=True, default=False, help="Show inactive project streams")
@click.option("--ofuri", "ofuri", type=click.STRING, shell_complete=get_product_stream_ofuris)
@click.pass_context
@progress_bar
def get_product_stream(ctx, product_stream_name, inactive, ofuri):
    """Retrieve Product Stream."""
    if not ofuri and not product_stream_name:
        click.echo(ctx.get_help())
        exit(0)
    session = CorgiService.create_session()
    cond = {}
    if ofuri:
        cond["ofuri"] = ofuri
    if product_stream_name:
        cond["name"] = product_stream_name
    data = session.product_streams.retrieve_list(**cond)
    return cprint(data, ctx=ctx)


@product_streams.command(name="latest-components")
@click.argument(
    "product_stream_name",
    required=False,
    type=click.STRING,
    shell_complete=get_product_stream_names,
)
@click.option("--namespace", type=click.Choice(CorgiService.get_component_namespaces()), help="")
@click.option("--ofuri", "ofuri", type=click.STRING, shell_complete=get_product_stream_ofuris)
@click.option("--view", default="summary")
@click.pass_context
def get_product_stream_components(ctx, product_stream_name, namespace, ofuri, view):
    """Retrieve Product Stream latest Components."""
    if not ofuri and not product_stream_name:
        click.echo(ctx.get_help())
        exit(0)
    ctx.invoke(list_components, ofuri=ofuri)


@product_streams.command(name="manifest")
@click.argument(
    "product_stream_name",
    required=False,
    type=click.STRING,
    shell_complete=get_product_stream_names,
)
@click.option("--ofuri", "ofuri", type=click.STRING, shell_complete=get_product_stream_ofuris)
@click.pass_context
def get_product_stream_manifest(ctx, product_stream_name, ofuri):
    """Retrieve Product Stream manifest."""
    if not ofuri and not product_stream_name:
        click.echo(ctx.get_help())
        exit(0)
    session = CorgiService.create_session()
    pv = None
    if ofuri:
        pv = session.product_streams.retrieve_list(ofuri=ofuri).additional_properties
    if product_stream_name:
        pv = session.product_streams.retrieve_list(name=product_stream_name).additional_properties
    if pv:
        data = session.product_streams.retrieve_manifest(pv["uuid"])
        return cprint(data, ctx=ctx)


@entities_grp.group(help=f"{CORGI_API_URL}/api/v1/builds")
@click.pass_context
def builds(ctx):
    pass


@builds.command(name="list")
@click.argument("software_build_name", required=False)
@click.pass_context
def list_software_builds(ctx, software_build_name):
    """Retrieve a list of Software Builds."""
    session = CorgiService.create_session()
    cond = default_conditions
    if software_build_name:
        cond["name"] = software_build_name
    data = session.builds.retrieve_list(**cond, limit=1000).results
    return cprint(data, ctx=ctx)


@builds.command(name="get")
@click.argument("build_id", required=False)
@click.pass_context
def get_software_builds(ctx, build_id):
    """Retrieve Software Build."""
    session = CorgiService.create_session()
    data = session.builds.retrieve(build_id)
    return cprint(data, ctx=ctx)
